import enum
from types import NoneType
from typing import Any, List, Optional, Union, get_type_hints
try:
    from typing import get_args, get_origin
except ImportError:
    get_args = lambda t: getattr(t, '__args__', ())
    get_origin = lambda t: getattr(t, '__origin__', None)

class _LoadableSwayObject:

    @staticmethod
    def _value(field_type, value):
        if value is None or value == "none":
            return None
        if issubclass(field_type, _LoadableSwayObject):
            return field_type.from_dict(value)
        return field_type(value)

    @classmethod
    def from_dict(cls, data:dict[str,Any]):
        self = cls()
        hints = get_type_hints(cls)
        for key, value in data.items():
            if key not in hints:
                setattr(self, key, value)
                continue
            field_type = hints[key]
            field_origin = get_origin(field_type)
            field_args = set(get_args(field_type))
            
            if field_origin == list:
                field_type = field_args.pop()
                setattr(self, key, [self._value(field_type, v) for v in value])
            
            elif field_origin == Union and NoneType in field_args and len(field_args) == 2:
                field_args.remove(NoneType)
                field_type = field_args.pop()
                setattr(self, key, self._value(field_type, value))
            
            else:
                setattr(self, key, self._value(field_type, value))
        return self

    def __repr__(self):
        kv = [
            f"{n}={repr(getattr(self, n))}" 
            for n in dir(self)
            if not n.startswith('_') and n not in ['from_dict']
        ]
        return f"{self.__class__.__name__}({', '.join(kv)})"




class CommandResult(_LoadableSwayObject):
    success:bool
    parse_error:Optional[bool]
    error:Optional[str]

class Gaps(_LoadableSwayObject):
    top:int
    left:int
    bottom:int
    right:int

class BarConfig(_LoadableSwayObject):
    id:str
    mode:str
    position:str
    status_command:str
    font:str
    workspace_buttons:bool
    workspace_min_width:int
    binding_mode_indicator:bool
    verbose:bool
    colors:dict
    gaps:Gaps
    bar_height:int
    status_padding:int
    status_edge_padding:int
 
class Version(_LoadableSwayObject):
    major:int
    minor:int
    patch:int
    human_readable:str
    loaded_config_file_name:str

class LibInput(_LoadableSwayObject):
    send_events:Optional[str]
    tap:Optional[str]
    tap_button_map:Optional[str]
    tap_drag:Optional[str]
    tap_drag_lock:Optional[str]
    accel_speed:Optional[float]
    accel_profile:Optional[str]
    natural_scroll:Optional[str]
    left_handed:Optional[str]
    click_method:Optional[str]
    middle_emulation:Optional[str]
    scroll_method:Optional[str]
    scroll_button:Optional[int]
    dwt:Optional[str]
    dwtp:Optional[str]
    calibration_matrix:Optional[list]
 
class Input(_LoadableSwayObject):
    identifier:str
    name:str
    vendor:str
    product:str
    type:str
    xkb_active_layout_name:Optional[str]
    xkb_layout_names:Optional[list]
    xkb_active_layout_index:Optional[str]
    scroll_factor:Optional[float]
    libinput:Optional[LibInput]
    repeat_delay:Optional[float]
    repeat_rate:Optional[float]
 
class Seat(_LoadableSwayObject):
    name:str
    capabilities:int
    focus:int
    devices:List[Input]

class Hinting(enum.Enum):
    RGB = 'rgb'
    BGR = 'bgr'
    VRGB = 'vrgb'
    VBGR = 'vbgr'
    unknown = "unknown"

class BorderStyle(enum.Enum):
    normal = "normal"
    pixel = "pixel"
    csd = "csd"

class LayoutType(enum.Enum):
    splith = "splith"
    splitv = "splitv"
    stacked = "stacked"
    tabbed = "tabbed"
    output = "output"

class Orientation(enum.Enum):
    vertical= 'vertical'
    horizontal = 'horizontal'

class X11Window(_LoadableSwayObject):
    title:str
    class_:str 
    instance:Any
    window_role:Any
    window_type:Any
    transient_for:Any

class Rect(_LoadableSwayObject):
    x:int
    y:int
    width:int
    height:int

class NodeType(enum.Enum):
    root = "root"
    output = "output"
    workspace = "workspace"
    con = "con"
    floating_on = "floating_con"

class Workspace(_LoadableSwayObject):
    id:int
    name:str
    rect:Rect
    visible:bool
    focused:bool
    urgent:bool
    num:int
    output:str

class OutputMode(_LoadableSwayObject):
    width:int
    height:int
    refresh:float

class OutputTransform(enum.Enum):
    NORMAL = 'normal'
    T_90 = '90'
    T_180 = '180'
    T_270 = '170'
    FLIPPED_90 = 'flipped-90'
    FLIPPED_180 = 'flipped-180'
    FLIPPED_270 = 'flipped-270'

class Output(_LoadableSwayObject):
    name:str
    rect:Rect
    make:str
    model:str
    serial:str
    active:bool
    dpms:bool
    power:bool
    primary:bool
    scale:float
    subpixel_hinting:Optional[Hinting]
    transform:OutputTransform
    current_workspace:Optional[str]
    modes:List[OutputMode]
    current_mode:OutputMode




class RootNode(_LoadableSwayObject):
    id:int
    name:str
    type:NodeType
    border:Optional[BorderStyle]
    current_border_width:int
    layout:LayoutType
    orientation:Optional[Orientation]
    percent:Optional[float]
    rect:Rect
    window_rect:Rect 
    deco_rect:Rect 
    geometry:Rect 
    urgent:bool
    sticky:bool
    marks:List[str]
    focused:bool
    focus:List[int]
    nodes:List['RootNode']
    floating_nodes:List['RootNode']

class WorkspaceNode(RootNode, Workspace):
    representation:str

class ContainerNode(RootNode):
    fullscreen_mode:int

class ViewNode(RootNode):
    app_id:Optional[str]
    pid:int
    visible:bool
    shell:str
    inhibit_idle:bool
    idle_inhibitors:Any

class X11ViewNode(ViewNode):
    window:Optional[int]
    window_properties: Optional[X11Window]

class OutputNode(RootNode, Output):...



from typing import Any, Literal, Optional, List, Final, Dict, Self, Protocol
import enum
from .ipc import EventType
from .model import _LoadableSwayObject, Input, Workspace, RootNode, BarConfig

class WorkspaceChangeType(enum.Enum):
    init = "init"
    empty = "empty"
    focus = "focus"
    move = "move"
    rename = "rename"
    urgent = "urgent"
    reload = "reload"

class WindowChangeType(enum.Enum):
    new = "new"
    close = "close"
    focus = "focus"
    title = "title"
    fullscreen_mode = "fullscreen_mode"
    move = "move"
    floating = "floating"
    urgent = "urgent"
    mark = "mark"

ALL_EVENT_TYPES = [ EventType.workspace, EventType.mode, EventType.window, EventType.barconfig_update, EventType.binding,
                    EventType.shutdown, EventType.tick, EventType.bar_state_update, EventType.input ]

class SwayEvent(_LoadableSwayObject):...

class WorkspaceEvent(SwayEvent):
    change: WorkspaceChangeType
    current: Workspace
    old: Optional[Workspace]

class ModeEvent(SwayEvent):
    change:str
    pango_markup:bool

class WindowEvent(SwayEvent):
    change:WindowChangeType
    container:RootNode

class BindingEvent(SwayEvent):
    change:str
    command:str
    event_state_mask:List[str]
    input_code:int
    symbol:str
    input_type:str

class ShutdownEvent(SwayEvent):
    change:Literal["exit"]

class TickEvent(SwayEvent):
    first:bool
    payload:str

class BarStateUpdateEvent(SwayEvent):
    id:str
    visible_by_modifier:bool

class InputEvent(SwayEvent):
    change:str
    input:Input

class BarConfigUpdateEvent(SwayEvent, BarConfig):...

class CanBeBuiltFromDict(Protocol):
    @classmethod
    def from_dict(cls, data:Dict[str, Any]) -> Self:...

EVENT_TYPE_TO_EVENT:Final[Dict[EventType, CanBeBuiltFromDict]] = {
    EventType.bar_state_update: BarStateUpdateEvent,
    EventType.barconfig_update: BarConfigUpdateEvent,
    EventType.binding: BindingEvent,
    EventType.input: InputEvent,
    EventType.mode: ModeEvent,
    EventType.tick: TickEvent,
    EventType.shutdown: ShutdownEvent,
    EventType.window: WindowEvent,
    EventType.workspace: WorkspaceEvent
}



import json
from typing import Any, Optional, Sequence, overload
from .event import EVENT_TYPE_TO_EVENT
from .ipc import send_ipc_message, PayloadType, EventType, get_ipc_socket, serialize_message, deserialize_message, PAYLOAD_MAGIC_STRING, RESPONSE_BUFFER_SIZE
from .model import CommandResult, Gaps, BarConfig, Version, LibInput, Input, Seat, Hinting,\
                   BorderStyle, LayoutType, Orientation, X11Window, Rect, NodeType, Workspace,\
                   OutputMode, OutputTransform, Output, RootNode, WorkspaceNode, ContainerNode, \
                   ViewNode, X11ViewNode, OutputNode

def run_command(command:str) -> list[CommandResult]:
    result = send_ipc_message(
            PayloadType.RUN_COMMAND, 
            command
        )
    return [CommandResult.from_dict(r) for r in result]

def get_workspaces() -> list[Workspace]:
    result = send_ipc_message(PayloadType.GET_WORKSPACES)
    return [Workspace.from_dict(op) for op in result]

def subscribe(events:Sequence[EventType]):
    """ Subscribe to the given sequence of events.
    
    This function returns a generator that yields each message recieved from the IPC socket.
    """
    with get_ipc_socket() as s:
        s.send(serialize_message(PayloadType.SUBSCRIBE, json.dumps([e.name for e in events])))
        buffer = b""
        while True:
            buffer += s.recv(RESPONSE_BUFFER_SIZE)
            while PAYLOAD_MAGIC_STRING in buffer:
                mtype, message, buffer = deserialize_message(buffer)
                if mtype == PayloadType.SUBSCRIBE: continue
                if not isinstance(mtype, EventType):
                    raise TypeError(f"The payload type of a subscribe response is not an event. mtype={mtype}")
                yield EVENT_TYPE_TO_EVENT[mtype].from_dict(message)

def get_outputs() -> list[Output]:
    result = send_ipc_message(PayloadType.GET_OUTPUTS)
    return [Output.from_dict(op) for op in result]

def get_tree() -> RootNode:
    return RootNode.from_dict(send_ipc_message(PayloadType.GET_TREE))

def get_marks() -> list[str]:
    return send_ipc_message(PayloadType.GET_MARKS)

@overload
def get_bar_config() -> list[str]:...
@overload
def get_bar_config(bar_id:str) -> BarConfig:...
def get_bar_config(bar_id:Optional[str]=None) -> Any:
    if bar_id is None:
        return send_ipc_message(PayloadType.GET_BAR_CONFIG)
    return BarConfig.from_dict(send_ipc_message(PayloadType.GET_BAR_CONFIG, bar_id))

def get_version() -> Version:
    return send_ipc_message(PayloadType.GET_VERSION)

def get_binding_modes() -> list[str]:
    return send_ipc_message(PayloadType.GET_BINDING_MODES)

def get_config() -> str:
    return send_ipc_message(PayloadType.GET_CONFIG)["config"]

def send_tick(payload:str="") -> bool:
    return send_ipc_message(PayloadType.SEND_TICK, payload)["success"]

def sync() -> bool:
    return send_ipc_message(PayloadType.SYNC)["success"]

def get_binding_state() -> str:
    return send_ipc_message(PayloadType.GET_BINDING_STATE)

def get_inputs() -> list[Input]:
    return [
        Input.from_dict(op)
        for op in send_ipc_message(PayloadType.GET_INPUTS)
    ]

def get_seats() -> list[Seat]:
    return [Seat.from_dict(op) for op in send_ipc_message(PayloadType.GET_SEATS)]

__ALL__ = [
    "run_command", "get_workspaces", "subscribe", "get_outputs", "get_tree", "get_marks",
    "get_bar_config", "get_version", "get_binding_modes", "get_config", "send_tick", "sync", 
    "get_binding_state", "get_inputs", "get_seats", 
    
    "CommandResult", "Gaps", "BarConfig", "Version", "LibInput", "Input", "Seat", "Hinting",
    "BorderStyle", "LayoutType", "Orientation", "X11Window", "Rect", "NodeType", "Workspace",
    "OutputMode", "OutputTransform", "Output", "RootNode", "WorkspaceNode", "ContainerNode", 
    "ViewNode", "X11ViewNode", "OutputNode"
]



"""
https://man.archlinux.org/man/sway-ipc.7.en

Payload structure

    <magic-string> <payload-length> <payload-type> <payload>

    Where,
        <magic-string>   = "i3-ipc"
        <payload-length> = 32-bit integer in native byte order
        <payload-type>   = 32-bit integer in native byte order 
        <payload>        = The bytes containing the payload

"""
import json
import socket
import sys
import os
import enum
from typing import Any, Final, Optional

PAYLOAD_MAGIC_STRING:Final = b"i3-ipc"
SWAY_SOCK_ENV_VAR:Final = 'I3SOCK'
RESPONSE_BUFFER_SIZE:Final = 1024 * 100

class PayloadType(enum.Enum):
    RUN_COMMAND = 0
    GET_WORKSPACES = 1
    SUBSCRIBE = 2
    GET_OUTPUTS = 3
    GET_TREE = 4
    GET_MARKS = 5
    GET_BAR_CONFIG = 6
    GET_VERSION = 7
    GET_BINDING_MODES = 8
    GET_CONFIG = 9
    SEND_TICK = 10
    SYNC = 11
    GET_BINDING_STATE = 12
    GET_INPUTS = 100
    GET_SEATS = 101

class EventType(enum.Enum):
    workspace = 0x80000000
    mode = 0x80000002
    window = 0x80000003
    barconfig_update = 0x80000004
    binding = 0x80000005
    shutdown = 0x80000006
    tick = 0x80000007
    bar_state_update = 0x80000014
    input = 0x80000015

def get_socket_location() -> str:
    """Obtain the Sway socket location via the I3SOCK environment variable."""
    if SWAY_SOCK_ENV_VAR in os.environ:
        return os.environ[SWAY_SOCK_ENV_VAR]
    raise ValueError('No default socket location available. Is Sway installed and running?')

def get_ipc_socket(socket_location:Optional[str]=None) -> socket.socket:
    """Get a Sway IPC socket."""
    if socket_location is None:
        socket_location = get_socket_location()
    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    s.connect(socket_location)
    return s

def serialize_message(payload_type:PayloadType, payload:str|bytes) -> bytes:
    """Take a payload type and payload body, and serialize it into a series of bytes in the expected format."""
    if isinstance(payload, str):
        payload = payload.encode('utf-8')
    result:list[bytes] = []
    result.append(PAYLOAD_MAGIC_STRING)
    result.append(len(payload).to_bytes(4, byteorder=sys.byteorder))
    result.append(payload_type.value.to_bytes(4, byteorder=sys.byteorder))
    result.append(payload)
    return b"".join(result)

def deserialize_message(payload:bytes) -> tuple[PayloadType|EventType, Any, bytes]:
    """Take a message recieved from the IPC socket, and parse it into a `Payload` object."""
    if not payload.startswith(PAYLOAD_MAGIC_STRING):
        raise Exception("Payload contents does not begin with magic string.")
    payload = payload[len(PAYLOAD_MAGIC_STRING):]
    payload_length = int.from_bytes(payload[:4], byteorder=sys.byteorder)
    payload_type_id = int.from_bytes(payload[4:8], byteorder=sys.byteorder)
    if payload_type_id > 101:
        payload_type = EventType(payload_type_id)
    else:
        payload_type = PayloadType(payload_type_id)
    current_payload = payload[8:8+payload_length]
    return payload_type, json.loads(current_payload.decode('utf-8')), payload[8+payload_length:]

def send_ipc_message(ptype: PayloadType, payload:Any="") -> Any:
    """Send a message to the Sway IPC consisting of the given payload type and message."""
    with get_ipc_socket() as s:
        s.send(serialize_message(ptype, payload))
        buffer = s.recv(RESPONSE_BUFFER_SIZE)
        mtype, message, remaining = deserialize_message(buffer)
        if mtype != ptype:
            raise Exception(f"IPC response type {mtype} does not match sent type {ptype}")
        if len(remaining) > 0:
            raise Exception(f"There should not be any buffer remainder from standard IPC responses. r={repr(remaining)}")
        return message
